import re
import numpy as np

class DatabaseTable:
    def __init__(self,name,params):
        self.name = name
        self.params = params
    
    def getColumns(self):
        return self.params['columns']

    def getRowClassifier(self):
        return self.params['classifier'].upper()
    
    def __hash__(self):
        return hash(name)
    
    def __eq__(self,other):
        return self.name == other.name

class DatabaseTemporalTable(DatabaseTable):
    def __init__(self,name,params):
        super().__init__(self,params)

    def getDateCol(self):
        return self.params['dateCol']

#this class keeps track of database metadata to make SQL queries built from the SQLTranslate module more specific
#also will reduce processing time for SQL Translate as it can access from a cache
class DatabaseInfo:
    def __init__(self, tableInfo, dataCategories=None):
        self.tables = {}
        for table in tableInfo:
            self.tables[table] = DatabaseTemporalTable(table,tableInfo[table]) if tableInfo[table]['temporal'] else DatabaseTable(table,tableInfo[table])
        self.dataCategories = dataCategories
        self.columnChoices = {}
        self.commonColumns = {}
        for table in tableInfo:
            self.generateColumnChoices(table)
        self.temporalDataTables = [table for table in tableInfo if tableInfo[table]['temporal']]

    def generateColumnChoices(self,table):
        for column in self.tables[table].getColumns():
            if(column in self.columnChoices):
                for otherTable in self.columnChoices[column]:
                    self.commonColumns[(table,otherTable)] = column
                self.columnChoices[column].append(table)
            else:
                self.columnChoices[column] = [table]

    def getColumnsAndTablesFromCategories(self,categories):
        columnList = []
        tableList = set()
        for category in categories:
            for table, items in self.dataCategories[category].items():
                tableList.add(table)
                columnList.extend(list(map(lambda columnName: table+"." + columnName.upper(), items)))
        tablePairs = [tablePair for tablePair in self.commonColumns.keys() if (tablePair[0] in tableList) and (tablePair[1] in tableList)]
        commonColumns = list(map(lambda tablePair: (tablePair[0] + "." + self.commonColumns[tablePair], tablePair[1] + "." + self.commonColumns[tablePair]), tablePairs))
        columnList = list(set(columnList))
        return {'columns': columnList,
                'tables': tableList,
                'commonColumns':commonColumns}

    def getColumnsAndTablesFromColumnList(self,columns):
        columnList = list(map(lambda columnName: self.columnChoices[columnName][0]+"."+columnName.upper(),columns))
        tableList = set([self.columnChoices[columnName][0] for columnName in columns])
        tablePairs = [tablePair for tablePair in self.commonColumns.keys() if (tablePair[0] in tableList) and (tablePair[1] in tableList)]
        commonColumns =  list(map(lambda tablePair: (tablePair[0] + "." + self.commonColumns[tablePair], tablePair[1] + "." + self.commonColumns[tablePair]), tablePairs))
        return {'columns': columnList,
                'tables': tableList,
                'commonColumns':commonColumns}

    def getCategories(self):
        return self.dataCategories

    def getRowClassifierColumns(self,tables):
        return list(map(lambda table: table+"."+ self.tables[table].getRowClassifier().upper(),tables))

    def getDateTimeColumns(self,tables):
        return list(map(lambda table: table+ "."+ self.tables[table].getDateCol().upper(),[t for t in self.tables if t in self.temporalDataTables]))
    




import os
from os import path
import sys
# import DatabaseInfo
from .DatabaseInfo import *
# import sqltranslate
from .sqltranslate import *
import json
import pickle


"""for running sql translate on shell through command line"""

def getDBObj():
    cacheOutOfDate = False
    if(os.path.exists("DatabaseInfoCache")):
        if(os.path.exists("tableInfo.json") and os.path.getmtime("tableInfo.json")>os.path.getmtime("DatabaseInfoCache")):
            cacheOutOfDate = True
        if(os.path.exists("dataCategories.json") and os.path.getmtime("dataCategories.json")>os.path.getmtime("DatabaseInfoCache")):
            cacheOutOfDate = True
    else:
        cacheOutOfDate = True

    if(cacheOutOfDate):
        if(os.path.exists("tableInfo.json")):
            with open("tableInfo.json","r") as f:
                try:
                    tableInfo = json.load(f)
                except json.decoder.JSONDecodeError:
                    print("improper file format '{}".format("tableInfo.json"),file=sys.stderr)
                    sys.exit()
            dataCategories = None
            if(os.path.exists("dataCategories.json")):
                with open("dataCategories.json","r") as f:
                    try:
                        dataCategories = json.load(f)
                    except json.decoder.JSONDecodeError:
                        print("improper file format '{}".format("dataCategories.json"),file=sys.stderr)
                        sys.exit()
        else:
            print("database info file does not exist '{}'".format("tableInfo.json"),file=sys.stderr)
            sys.exit()
        dbInfo = DatabaseInfo(tableInfo,dataCategories)
        with open("DatabaseInfoCache","wb") as f:
            pickle.dump(dbInfo,f,protocol = pickle.HIGHEST_PROTOCOL)

    if(not cacheOutOfDate):
        with open("DatabaseInfoCache","rb") as f:
            f.seek(0)
            dbInfo = pickle.load(f)
    return dbInfo


def getParams(databaseInfo):
    number = re.compile('\\d+')
    params = {}
    temporalColumnChoices = [column for column in databaseInfo.columnChoices if any(table in databaseInfo.temporalDataTables for table in databaseInfo.columnChoices[column])]
    temporalDataCategories = [category for category in databaseInfo.dataCategories if any(column in list(databaseInfo.dataCategories[category].values())[0] for column in temporalColumnChoices)]
    nonTemporalColumnChoices = list(set(databaseInfo.columnChoices.keys()) - set(temporalColumnChoices))
    nonTemporalDataCategories = list(set(databaseInfo.dataCategories.keys()) - set(temporalDataCategories))

    temporal = True if input("temporal data? (y/n)")=="y" else False
    categorical = True if input("choose data from categories or columns? (cat/col)")=='cat' else False
    params['categorical']=categorical
    filters = {}
    if(temporal):
        if(categorical):
            print(temporalDataCategories)
            selected = input("choose from these categories, separate by commas ").strip().split(',')
            params['categories'] = selected
        else:
            print(temporalColumnChoices)
            selected = input("choose from columns, separate by commas ").strip().split(',')
            params['columnList'] = selected
            print("select filter type and values for each column. press enter for none:")
            for selection in selected:
                filterType = input(selection + ": filter by range, equality, or none (press enter)? (r/e)").strip()
                if(len(filterType)):
                    if(filterType=='r'):
                        values = input("enter lower and upper bound of range (inclusive) separated by comma: ").strip().split(',')
                    else:
                        values = input("enter values you want " + selection + " to be equal to separated by commas ").strip().split(',')
                    if(re.match(number,values[0])!=None):
                        values = [int(value) for value in values]
                    filters[selection] = {'type': 'range' if filterType=='r' else 'equal','values':values}
        
        aggregateBy = input("how frequently to aggregate data? (year/month/day) ").strip()
        aggregate = {}
        if(aggregateBy!="day"):
            print("select aggregation functions for each selection (avg/sum/med):")
            for selection in selected:
                method = input(selection + ": ")
                aggregate[selection] = method
        
        dateRange = input("select date range (lower and upper bound separated by comma, enter for none): ").strip()
        if(len(dateRange)):
            params['dateRange'] = dateRange.split(',')
        params['aggregateBy'] = aggregateBy
        params['aggregate'] = aggregate
    else:
        if(categorical):
            print(nonTemporalDataCategories)
            selected = input("choose from these categories, separate by commas ").strip().split(',')
            params['categories'] = selected
        else:
            print(nonTemporalColumnChoices)
            selected = input("choose from columns, separate by commas ").strip().split(',')
            params['columnList'] = selected
            for selection in selected:
                filterType = input(selection + ": filter by range, equality, or none (press enter)? (r/e)").strip()
                if(len(filterType)):
                    if(filterType=='r'):
                        values = input("enter lower and upper bound of range (inclusive) separated by comma: ").strip().split(',')
                    else:
                        values = input("enter values you want " + selection + " to be equal to separated by commas ").strip().split(',')
                    if(re.match(number,values[0])!=None):
                        values = [int(value) for value in values]
                    filter[selection] = values
        if(input("choose aggregate functions for selected? (y/n)")=="y"):
            aggregate = {}
            print(selection + ": select aggregation functions for each selection (avg/sum/med):")
            for selection in selected:
                method = input(selection + ": ").strip()
                if(len(method)):
                    aggregate[selection] = method
            if(len(aggregate)):
                params['aggregate'] = aggregate

    params['filters'] = filters
    return params
    

if __name__=='__main__':
    dbInfo = getDBObj()
    params = getParams(dbInfo)
    print(params)
    if('dateRange' in params):
        SQLTran = SQLTranslateTemporal(dbInfo,params)
    elif('aggregate' in params and len(params['aggregate'])):
        SQLTran = SQLTranslateAggregate(dbInfo,params)
    else:
        SQLTran = SQLTranslate(dbInfo,params)
    print(SQLTran)


"""Main module."""
from itertools import repeat, combinations
from .DatabaseInfo import *
from .__main__ import *



class SQLTranslate:
    def __init__(self,databaseInfo,params):
        self.params = params
        self.databaseInfo = databaseInfo
        self.selected = databaseInfo.getColumnsAndTablesFromCategories(self.params['categories']) if self.params['categorical'] else databaseInfo.getColumnsAndTablesFromColumnList(self.params['columnList'])
        self.selectedColumns = self.selected['columns']
        self.selectedTables = self.selected['tables']
        self.selectedColumns.extend(self.databaseInfo.getRowClassifierColumns(self.selectedTables))
        self.join = self.selected['commonColumns']
        self.filters = []
        if(not self.params['categorical'] and 'filters' in self.params):
            colSet = set()
            for i in range(len(self.selectedColumns)):
                currCol = self.selectedColumns[i].split('.')[1].lower()
                if(currCol in self.params['filters'] and not currCol in colSet):
                    colSet.add(currCol)
                    currFilt = self.params['filters'][currCol]
                    if(currFilt['type']=='range'):
                        self.filters.append(("(" + self.selectedColumns[i] + " BETWEEN " + "\"" + currFilt['values'][0] + "\" AND " + "\"" + currFilt['values'][1] + "\")") if isinstance(currFilt['values'][0],str) else self.filters.append("(" + self.selectedColumns[i] + " BETWEEN " +  str(currFilt['values'][0]) + " AND " + str(currFilt['values'][1]) + ")"))
                    else:
                        self.filters.extend(list(map(lambda filt: "(" + selectedColumns[i] + "=" + "\"" + filt + "\")", currFilt['values']))) if isinstance(currFilt['values'][0],str) else self.filters.extend(list(map(lambda filt: "(" + self.selectedColumns[i] + "=" +  str(filt) + ")", currFilt['values'])))

    def select(self):
        return "SELECT " + ', '.join(list(set(self.selectedColumns)))
    
    def fromTables(self):
        return "FROM " + ' inner join '.join(list(map(lambda t: "\"" + t.upper() + "\" as " + t, self.selectedTables)))
    
    def on(self):
        if(len(self.join)):
            return "on " + ' and '.join(list(map(lambda pair: pair[0] + "=" + pair[1],self.join)))
        return ""

    def where(self):
        if(len(self.filters)):
            return "WHERE " + ' and '.join([filter for filter in self.filters if filter!=None])
        return ""

    def command(self):
        return '\n'.join([string for string in [self.select(),self.fromTables(),self.on(),self.where()] if len(string)])

    def __str__(self):
        return self.command()

class SQLTranslateAggregate(SQLTranslate):
    def __init__(self,databaseInfo,params):
        super().__init__(databaseInfo,params)
        self.groupByCols = []
        self.aggregate = []
        if(self.params['categorical']):
            for category in self.params['categories']:
                categoryDict = self.databaseInfo.dataCategories[category]
                self.aggregate.extend([self.params['aggregate'][category]]*len(list(categoryDict.values())[0]))
            self.groupByCols.extend([self.selectedColumns[i] for i in range(len(self.aggregate),len(self.selectedColumns))])
            self.aggregate.extend([None]*(len(self.selectedColumns)-len(self.aggregate)))
        else:
            for column in self.selectedColumns:
                currCol = column.split('.')[1].lower()
                if(currCol in self.params['aggregate']):
                    self.aggregate.append(self.params['aggregate'][currCol])
                else:
                    self.aggregate.append(None)
                    self.groupByCols.append(column)
        self.selectedColumns = list(map(lambda i: self.aggregate[i].upper() + "(" + self.selectedColumns[i] + ")" if self.aggregate[i]!=None else self.selectedColumns[i],[i for i in range(len(self.selectedColumns))]))
    
    def groupBy(self):
        return "GROUP BY " + ", ".join(list(set(self.groupByCols) ))

    def command(self):
        return '\n'.join(string for string in [super().command(), self.groupBy()] if len(string))

class SQLTranslateTemporal(SQLTranslateAggregate):
    def __init__(self,databaseInfo,params):
        self.aggregateBy = params['aggregateBy']
        self.dateRange = list(map(lambda date: "\"" + date + "\"", params['dateRange'])) if self.aggregateBy == 'day' else list(map(lambda date: self.aggregateBy.upper() + "(\"" + date +"\")",params['dateRange']))
        super().__init__(databaseInfo,params)
    
    def select(self):
        dateCol = self.databaseInfo.getDateTimeColumns(self.selectedTables)[0]
        if(self.aggregateBy=='month'):
            self.selectedColumns.extend(["MONTH("+dateCol+")","YEAR("+dateCol+")"])
        elif(self.aggregateBy=='year'):
            self.selectedColumns.append("YEAR("+dateCol+")")
        else:
            self.selectedColumns.append(dateCol)
        return super().select()

    def on(self):
        dateCols = self.databaseInfo.getDateTimeColumns(self.selectedTables)
        datePairs = list(combinations(dateCols,2))
        self.join.extend(datePairs)
        return super().on()

    def where(self):
        dateCol = self.databaseInfo.getDateTimeColumns(self.selectedTables)[0]
        if('dateRange' in self.params and len(self.params['dateRange'])):
            self.filters.append(f"({dateCol} BETWEEN DATE(\"{self.params['dateRange'][0]}\") AND DATE(\"{self.params['dateRange'][1]}\"))")
        return super().where()

    def groupBy(self):
        dateCol = self.databaseInfo.getDateTimeColumns(self.selectedTables)[0]
        if(self.aggregateBy=='month'):
            self.groupByCols.extend(["MONTH("+dateCol+")","YEAR("+dateCol+")"])
        elif(self.aggregateBy=='year'):
            self.groupByCols.append("YEAR("+dateCol+")")
        else:
            self.groupByCols.append(dateCol)
        return super().groupBy()

def run(params):
    dbInfo = getDBObj()
    if('dateRange' in params):
        SQLTran = SQLTranslateTemporal(dbInfo,params)
    elif('aggregate' in params and len(params['aggregate'])):
        SQLTran = SQLTranslateAggregate(dbInfo,params)
    else:
        SQLTran = SQLTranslate(dbInfo,params)
    return SQLTran


#!/usr/bin/env python

"""Tests for `sqltranslate` package."""

import pytest
from click.testing import CliRunner
from sqltranslate.sqltranslate import *
from sqltranslate import cli
from sqltranslate.DatabaseInfo import *
#from sqltranslate.__main__ import *

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string

HDDatabase = None

def test_databaseinfo_constructor():
    tableInfo = {
        'climatic':                                       
        {                                                 
            'temporal': True,
            'dateCol': 'day',                          
            'classifier': 'staid',
            'columns': ['tmmx','tmmn','pr','pet','vpd',
            'rmax','rmin','th','vs','day','staid']
        },
        'streamflow':                                     
        {                                                 
            'temporal': True,
            'dateCol': 'date',                             
            'classifier':'staid',
            'columns':['streamflow','date','staid']
        },
        'stations':                                       
        {
            'temporal': False,
            'classifier':'staid',
            'columns':['staid','lat','lng']
        }
    }
    dataCategories = {
        'Streamflow': {
           'streamflow':['streamflow']
        },
        'Temperature':{
               'climatic': ['tmmx','tmmn']
        },
        'Precipitation':{
            'climatic': ['pr']
        },
        'Humidity': {
            'climatic': ['pet', 'vpd','rmax','rmin']
        },
        'Wind':{
           'climatic': ['th','vs']
        }
    }
    global HDDatabase
    HDDatabase = DatabaseInfo(tableInfo,dataCategories)
    

# def test_databaseinfo_categories():
#     global HDDatabase
#     result = HDDatabase.getColumnsAndTablesFromCategories(['Temperature','Streamflow'])
    
# def test_databaseinfo_columnlist():
#     global HDDatabase
#     result = HDDatabase.getColumnsAndTablesFromColumnList(['tmmx','tmmn','streamflow','pet'])

# def test_databaseinfo_rowclassifiers():
#     global HDDatabase
#     result = HDDatabase.getRowClassifierColumns(['climatic','streamflow'])

# def test_databaseinfo_datetime():
#     global HDDatabase
#     result = HDDatabase.getDateTimeColumns(['climatic','streamflow','stations'])

def test_sqltranslate():
    global HDDatabase
    param1 = {
        'categorical':False,
        'columnList':['tmmx','tmmn','streamflow','pet'],
        'filters': {
            'tmmx': {
                'type':'range',
                'values': [150,160]
            },
            'staid': {
                'type':'equal',
                'values':[1,2]
            }
        }
    }
    SQLTran1 = SQLTranslate(HDDatabase,param1)
    param2 = {
        'categorical':True,
        'categories':["Temperature","Streamflow"],
    }
    SQLTran2 = SQLTranslate(HDDatabase,param2)
    param3 = {
        'categorical':False,
        'columnList':['lat','lng']
    }
    SQLTran3 = SQLTranslate(HDDatabase,param3)

def test_sqltranslate_agg():
    global HDDatabase
    param1 = {'categorical':False,
    'columnList':['tmmx','tmmn','streamflow','pet'],
    'aggregate':{'tmmx':'avg','tmmn':'avg','streamflow':'sum','pet':'avg'},
    'filters':{}
    }
    SQLTran1 = SQLTranslateAggregate(HDDatabase,param1)
    param2 =  { 
    'categorical':True,
    'categories':['Temperature','Streamflow','Humidity'],
    'aggregate':{'Temperature':'avg','Streamflow':'sum','Humidity':'avg'},
    'filters':{}
    }
    SQLTran2 = SQLTranslateAggregate(HDDatabase,param2)
    
def test_sqltranslate_tmp():
    global HDDatabase 
    param1 = {   'categorical':False,
    'columnList':['tmmx','tmmn','streamflow','pet'],
    'aggregate':{'tmmx':'avg','tmmn':'avg','streamflow':'sum','pet':'avg'},
    'dateRange':['01-01-2004','01-01-2006'],
    'aggregateBy':'year'
    }
    SQLTran1 = SQLTranslateTemporal(HDDatabase, param1)
    param2 = {   'categorical':False,
    'columnList':['tmmx','tmmn','streamflow','pet'],
    'aggregate':{'tmmx':'avg','tmmn':'avg','streamflow':'sum','pet':'avg'},
    'dateRange':['01-01-2004','01-01-2006'],
    'aggregateBy':'month'
    }
    SQLTran2 = SQLTranslateTemporal(HDDatabase, param2)
    param3 = {   'categorical':False,
    'columnList':['tmmx','tmmn','streamflow','pet'],
    'filters':{'staid': {'type':'equal', 'values':[1,2]}},
    'aggregate':{'tmmx':'avg','tmmn':'avg','streamflow':'sum','pet':'avg'},
    'dateRange':['01-01-2004','01-01-2006'],
    'aggregateBy':'day'
    }
    SQLTran3 = SQLTranslateTemporal(HDDatabase, param3)
    print(SQLTran3)
    param4 = {   'categorical':True,
    'categories': ['Temperature','Streamflow'],
    'aggregate':{'Temperature':'avg','Streamflow':'sum'},
    'dateRange':['01-01-2004','01-01-2006'],
    'aggregateBy':'year'
    }
    SQLTran4 = SQLTranslateTemporal(HDDatabase, param4)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=7.0', 'numpy']

test_requirements = ['pytest>=3', ]

setup(
    author="Joy He",
    author_email='jh4337@columbia.edu',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="SQL Translate stores database info and generates complex sql queries so people who don't know sql can access the data they need in a user friendly way",
    entry_points={
        'console_scripts': [
            'sqltranslate=sqltranslate.__main__:main',
        ],
    },
    install_requires=requirements,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='sqltranslate',
    name='sqltranslate',
    packages=find_packages(include=['sqltranslate', 'sqltranslate.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/joyhe208/pkg_sqltranslate',
    version='0.1.2',
    zip_safe=False,
    license='MIT',
    download_url='https://github.com/joyhe208/pkg_sqltranslate/archive/refs/tags/0.1.2.tar.gz'
)



"""Top-level package for SQLTranslate."""
from sqltranslate.DatabaseInfo import *
from sqltranslate.sqltranslate import *
__author__ = """Joy He"""
__email__ = 'jh4337@columbia.edu'
__version__ = '0.1.0'
